## Processing imports
import numpy as np

## plotting imports
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.colors import LogNorm
from astropy.visualization import astropy_mpl_style

__all__ = ['plt_eye', 'plt_stars', 'plt_fits', 'plt_flat', 'plt_bias']

def plt_eye(data, num_sigma = 3):
    """
    plt_eye takes image data for plotting and computes the optimal colourbar range.

    Parameters
    ----------
    data: image/array
            image data in 2-dimensional array format.
    num_sigma: float
            Positive float value that defines how many standard deviations are to be considered when computing the range.
    Returns
    -------
    mean_up, mean_down: float
           Upper and lower colourbar limits for plotting.
    """
    mean = np.mean(data)
    std = np.std(data)
    mean_up = mean + num_sigma * std
    mean_down = mean - num_sigma * std

    return mean_up, mean_down

def plt_stars(data, x, y, r, num_sigma = 3):
        plt.style.use(astropy_mpl_style)
        plt.figure()
        # plt.imshow(data, cmap='viridis', vmin=0)
        up, down = plt_eye(data, num_sigma=num_sigma)
        plt.imshow(data, cmap='viridis', vmin=down, vmax=up)
        cbar = plt.colorbar()
        plt.grid(False)
        
        # for i in range(len(x)):
        #         circlei=plt.Circle((x[i],y[i]), r[i], edgecolor=, alpha = 0.75, linewidth = 1)
        #         plt.gcf().gca().add_artist(circlei)
        colours = ['r', 'g', 'b', 'y', 'cyan', 'w', 'm']
        plt.scatter(x, y, s = r ** 2, edgecolors = colours, alpha = 0.2) 
        for p in range(len(x)):
                plt.text(x[p], y[p], str(p))

        plt.show()

def plt_fits(image_data, cmap_str = 'viridis', num_sigma = 3):
    plt.style.use(astropy_mpl_style)
    #print(image_data.shape)

    plt.figure()
    #plt.imshow(np.transpose(np.fliplr(image_data)), cmap='viridis')
    #plt.imshow(image_data, cmap='viridis', norm=LogNorm())
    # plt.imshow(image_data, cmap=cmap_str, vmin=0)
    # edited
    up, down = plt_eye(image_data, num_sigma = num_sigma)
    plt.imshow(image_data, cmap=cmap_str, vmin=down, vmax=up)
    cbar = plt.colorbar() #.ColorbarBase(ax, cmap=cmap_str,norm=mpl.colors.Normalize(vmin=down, vmax=up))
    #cbar.set_clim(down, up)

    # .ColorbarBase(ax, cmap=cm,norm=mpl.colors.Normalize(vmin=-0.5, vmax=1.5))
    #
    plt.grid(False)

    plt.show()
    
def plt_flat(image_data, cmap_str = 'viridis'):
    plt.style.use(astropy_mpl_style)
    #print(image_data.shape)

    plt.figure()
    #plt.imshow(np.transpose(np.fliplr(image_data)), cmap='viridis')
    #plt.imshow(image_data, cmap='viridis', norm=LogNorm())
    plt.imshow(image_data, cmap=cmap_str, norm=LogNorm(vmin=image_data.min(), vmax=image_data.max()))
    plt.colorbar()
    plt.grid(False)

    plt.show()
    
def plt_bias(image_data, cmap_str = 'viridis'):

    plt.style.use(astropy_mpl_style)
    #print(image_data.shape)

    plt.figure()
    #plt.imshow(np.transpose(np.fliplr(image_data)), cmap='viridis')
    #plt.imshow(image_data, cmap='viridis', norm=LogNorm())
    plt.imshow(image_data, cmap=cmap_str, vmin=0)
    plt.colorbar()
    plt.grid(False)

    plt.show()


# def canyon (curve analysis @ York )
# remove uneeded functions


import dorado


import astroalign as aa
# from donuts import Donuts

## Organize and prepare your data
# Get a fits file formated date string for the current observing night.
night = dorado.get_night() 
print(night)
# Your data directory
# Set night as directory 
directory = 'C:/Data/' + night

# You can name your target object
target = 'dorado'

# Create a working directory
dorado.mkwrkdir(directory)


# Catalogue input data from the data directory
bias_list, flats_list, lights_list = dorado.checkdir(directory)





## Calibrate your data
# Read the data in
data_series = dorado.get_series(directory, lights_list)
flats_series = dorado.get_series(directory, flats_list)
bias_series = dorado.get_series(directory, bias_list)

# Produce master reduction images
bias = dorado.mastBias(directory, bias_list)
flat = dorado.mastFlat(directory, flats_list, bias)

# Calibrate data series
series = dorado.reduce_series(directory, lights_list, flat, bias, target)

# Write calibrated data
dorado.write_series(directory, series, target)





## Plate solve your data series
# Pick image to solve and store its data
data = series[0].data
# Set up a file string for dorado to locate image
expath = '/wrk/calibrated/' 
im = target +  '_0.calibrated.fit'
image_file_path = dorado.file_string(directory + expath, im)
# Pass the data and file string to dorado
solved = dorado.plate_solve(data, image_file_path, write_fits = True, write_name = target + '_solved')




# Align each image with the plate solved image
aa_series = []
for image in series:
    img_r, footprint_r = aa.register(image.data, solved.data)
    aa_series.append(img_r)

dorado.write_series(directory + '/wrk/aligned/', aa_series, target + '_a')












# # Align each image with the solved image
# p, (pos_0, pos_1) = astroalign.find_transform(series[0].data, series[20].data)

# for (x1, y1), (x2, y2) in zip(pos_0, pos_1):
#     print("({:.2f}, {:.2f}) in source --> ({:.2f}, {:.2f}) in target"
#           .format(x1, y1, x2, y2))

# # Set up a reference image
# refim = lights_list[0]
# d = Donuts(refimage = refim, image_ext = 0, subtract_bkg = True, ntiles = 64)
# # for each image, compute the x/y translation required to align the images onto the reference image
# xp, yp = [], []
# for image in lights_list:
#     shift_result = d.measure_shift(image)
#     x, y = shift_result.x, shift_result.y
#     # y = shift_result.y
#     xp.append(x.value)
#     yp.append(y.value)
#     # Also check out shift_result.sky_background
#     # print(x, y)


import numpy as np
'''
Fournax is an abbreviation of Fourier numerical astronomy extension, its name is a backronym styled to match the constellation 'fornax'. 

A major future update will bring a more powerful curve fitting function to this sub-package.
'''

# __all__ = ['four_curve', 'pour', 'var_ephem']
class Fournax():

        def four_curve(y, terms):
                """
                Four_curve takes a timeseries and 'smoothes' out higher order terms in frequency space and returns the smoothed values.

                Parameters
                ----------
                y: list or array
                        y values of a timeseries. Time is x.
                terms: int
                        terms to retain in fourier series.
                Returns
                -------
                Fcurve: list or array
                        Fourier smoothed y values
                """
                # Fourier series domain
                # Compute real valued Fourier transform
                f = np.fft.rfft(y)
                # Null or zero coefficients above ammount of series "terms"
                # This corresponds to undesired high-frequency terms
                f[terms+1:] = 0
                # Collapse back into function space, result is smoothed Fourier curve
                Fcurve = np.fft.irfft(f)
                
                return Fcurve

        def pour(y):
                '''
                pour takes a timeseries and computes the power in frequency space.

                Parameters
                ----------
                y: list or array
                        y values of a timeseries. Time is x.

                Returns
                -------
                p: list or array
                        power of y

                '''
                # Compute real valued Fourier transform
                f = np.fft.fft(y)
                # p = np.square(np.abs(f))
                p = np.square(np.abs(f))
                
                return p
        
        def var_ephem(OBS, epoch, period):
                '''
                var_ephem takes observed time(s) of max light for a repeating variable star and ephemeris data and returns O-C values as well as the corresponding cycle.

                Parameters
                ----------
                OBS: float, list, or array
                        Observed time(s) of max light
                epoch: float
                        Epoch of ephemeris
                period: float
                        Period of ephemeris 
                Returns
                -------
                cycle: float, list, or array
                        The cycle corresponding to the time(s) of max light
                OmC: float, list, or array
                        The O-C value for the time(s) of max light
                '''
                cycle_ref = (OBS-epoch)/period
                cycle = np.round(cycle_ref)
                OmC = OBS-(epoch+cycle*period)
                return cycle, OmC
        # def powerranger
# def maxlight
# def properties




import warnings
warnings.filterwarnings('ignore')

## file imports
import os
from astropy.io import fits 
from astropy.nddata import CCDData
from astropy.table import Table
from astropy.stats import mad_std

## Processing imports
import numpy as np
import ccdproc
from scipy.ndimage import zoom

## photometry imports
from astropy import units as u

# __all__ = ['reduce_series', 'mastFlat', 'mastBias', 'theMask', 'sky_est', 'series_arith', 'norm_flat', 'stack_im']

def reduce_series(imdir, imlist, flat, bias, target, resize = '', caldir = ''):
        # directory, night, imlist, flat, bias, vstar, expath = '', resize = ''
        """
        plate_solve takes fits image file data and the corresponding file string to the data and calls nova.astrometry.net to obtain and then integrate WCS into the HDU.

        Parameters
        ----------
        imdir: filestring
                Path to the image data directory.
        caldir: filestring
                Path to store the calibrated image data.
        imlist: array[string]
                array of filestrings of data to reduce.
        flat: CCDdata
                CCDdata of the flatfield.
        bias: CCDdata
                CCDdata of the bias.
        target: string
                Name of target to use in file output name.
        resize: float
                scale factor to resize image data. default is None
        caldir: filestring
                Path to store the calibrated image data.
            
        Returns
        -------
        series: array[CCDdata]
                The CCDdata array containing the reduced images.
        """
        stardir = os.getcwd()
        # imdir = directory + '/' + night + '/' + expath
        # caldir = directory + night + '/wrk/calibrated/'
        if caldir == '':
                caldir = imdir + '/wrk/calibrated/'
        os.chdir(caldir)
        series = []
        for i in range(len(imlist)):
                os.chdir(imdir)
                hdu = CCDData.read(imlist[i], unit=u.adu)
                hdu = ccdproc.ccd_process(hdu, master_bias = bias, master_flat = flat)
                hdu = ccdproc.cosmicray_lacosmic(hdu, sigclip=5)
                hdu.header['bias corrected'] = True
                hdu.header['flat corrected'] = True
                hdu.header['cosmicray corrected'] = True
                hdu.header['calibrated'] = True
                # may want to move this to before reduction step
                if (resize != ''):
                                hdu.header['Resized'] = True
                                hdu.data = zoom(hdu.data, (resize, resize), order=0)
                os.chdir(caldir)
                # mod to use imlist imlist[i]
                hdu.write(target + '-' + str(i) + '_c.fit')
                series.append(hdu)
        # print(len(series))
        os.chdir(stardir)

        return series

def mastFlat(directory, flats, bias, caldir = ''):
        """
        mastFlat takes a datadirectory string, a list of flats and a master bias image to construct 
        a master flatfield image.

        Parameters
        ----------
        directory: filestring
                Path to the image data directory.
        flats: array[CCDdata]
                array of raw flatfields.
        bias: CCDdata
                CCDdata of the bias.
        caldir: filestring
                Path to store the calibrated image data.

        Returns
        -------
        master_flat: CCDdata
                The combined master flatfield image.
        """
        # give option to median stack
        # give option to specify a filter for output.
        # directory, night, flats, bias
        # Allow resizing
        stardir = os.getcwd()

        path = directory # + '/' + night + '/'
        os.chdir(path)

        master_flat = ccdproc.combine(flats, method='average',
                             sigma_clip=True, sigma_clip_low_thresh=5, sigma_clip_high_thresh=5,
                             sigma_clip_func=np.ma.median, sigma_clip_dev_func=mad_std, unit=u.adu)
        master_flat.header['stacked'] = True

        if caldir == '':
                caldir = directory + '/wrk/flats/'
        os.chdir(caldir)
        master_flat.write('master_flat.fit')

        os.chdir(stardir)

        return master_flat

def mastBias(directory, bias, caldir = ''):
        """
        mastBias takes a datadirectory string, a list of bias images to construct 
        a master bias image.

        Parameters
        ----------
        directory: filestring
                Path to the image data directory.
        bias: array[CCDdata]
                array of raw bias images.
        caldir: filestring
                Path to store the calibrated image data.

        Returns
        -------
        master_bias: CCDdata
                The combined master bias image.
        """
        # directory, night, bias
        # Allow resizing
        # Allow specification of median or mean
        stardir = os.getcwd()

        path = directory # + '/' + night + '/'
        os.chdir(path)

        master_bias = ccdproc.combine(bias, method='average', unit=u.adu)
        master_bias.meta['stacked'] = True

        if caldir == '':
                caldir = directory +  '/wrk/bias/'
        os.chdir(caldir)
        master_bias.write('master_bias.fit')

        os.chdir(stardir)

        return master_bias

def theMask(data, lx, hx, ly, hy):
        """
        theMask takes an image and bound parameters to mask the outter edges of 
        the image in the case of imperfections. Data suffering from descecant 
        failure can be made usable via this.

        Parameters
        ----------
        data: CCDdata
                CCDdata of the image to be masked.
        lx, hx, ly, hy: integers
                The low and high x coordinates and low and high y coordinates
                respectively.
        Returns
        -------
        mask: array[Boolean]
                The masked array for application to the data image in further functions.
                
        """
        # Allow for multiple mask rectangles or circles
        # Star masking
        mask = np.zeros(data.shape, dtype=bool)
        mask[lx:hx, ly:hy] = True
        return mask

def sky_est(im):
        """
        sky_est makes a crude estimate of an images sky value based on a median
        method.

        Parameters
        ----------
        im: CCDdata
                CCDdata of the image to be solved.

        Returns
        -------
        sky: float
                image sky value.
        """
        # Allow for other sky definitions including masking
        # Use one from Proj1
        sky = np.median(im)

        return sky

def series_arith(series, operator, operand):
        """
        series_arith allows one to perform mathematical(algebraic) on a series of images.

        Parameters
        ----------
        series: array[CCDdata]
                Array of CCDdata of the image(s) to perform operations on.
        operator: string
                string of the algebaric operator.
        operand: CCDdata
                image to be used as an operand.

        Returns
        -------
        series: array[CCDdata]
                Array of CCDdata .
        """
        # mod to check datatype using type()
        # mod to remove im_count and make possible to use single image
        # mod to accomodate CCDdata object
        for i in range(len(series[0,0,:])):
            if (operator == '+'):
                series[:, :, i] = series[:, :, i]  + operand
            elif (operator == '-'):
                series[:, :, i] = series[:, :, i] - operand
            elif (operator == '/'):
                series[:, :, i] = series[:, :, i]  / operand
            elif (operator == '*'):
                series[:, :, i] = series[:, :, i]  * operand
        
        return series

def norm_flat(flat):
        """
            norm_flat takes a flatfield image and normalizes it for use in reduction.

            Parameters
            ----------
            flat: CCDdata
                CCDdata of the flatfield.

            Returns
            -------
            normFlat: CCDdata
                    The CCDdata object of the normalized flat.
        """
        # fix series_arith
        # beta test this function
        norm_factor = np.mode(flat, axis = None)[0]
        print("norm_factor: ", norm_factor)
        normFlat = series_arith([flat], '/', norm_factor)

        return normFlat

def stack_im(im_list):
        """
            stack_im stacks a series of images into a single frame.

            Parameters
            ----------
            imlist: array[string]
                array of filestrings of data to stack.

            Returns
            -------
            im: CCDdata
                    The stacked CCDdata image.
        """
        # Is there another version of this somewhere?
        # update for CCDdata
        # Add different methods for stacking (im combine)
        im = im_list[0]
        for i in range(1, len(im_list)):
            im = im + im_list[i]
        
        return im